import torch
import torch.nn as nn
import torch.nn.functional as F

num_c1 = 32
num_c2 = 64
num_c3 = 128

class AE(nn.Module):
    def __init__(self):
        super(AE, self).__init__()

        # encoder
        self.conv1 = nn.Conv2d(1, num_c1, 5, stride = 2, padding = 2, bias = True)
        self.conv2 = nn.Conv2d(num_c1, num_c2, 5, stride = 2, padding = 2, bias = True)
        self.conv3 = nn.Conv2d(num_c2, num_c3, 3, stride = 2, bias = True)
        self.fc1 = nn.Linear(num_c3 * 3 * 3, 10, bias = True)

        # decoder
        self.fc4 = nn.Linear(10, 3 * 3 * num_c3, bias = True)
        self.deconv3 = nn.ConvTranspose2d(num_c3, num_c2, 3, stride = 2, bias = True)
        self.deconv1 = nn.ConvTranspose2d(num_c2, num_c1, 5, stride = 2, padding=2, output_padding=1, bias = True)
        self.deconv2 = nn.ConvTranspose2d(num_c1, 1, 5, stride = 2, padding=2, output_padding=1, bias = True)

    def forward(self, x):

        # encoder
        #x = F.pad(x, (2, 1, 2, 1))
        x = F.relu(self.conv1(x))
        x = F.relu(self.conv2(x))
        x = F.relu(self.conv3(x))
        x = x.view(-1, num_c3 * 3 * 3)
        mid = self.fc1(x)
        norm = torch.sqrt(torch.sum(torch.pow(mid, 2), dim = 1))
        h = (mid.t()/norm).t()
        
        # decoder
        z = F.relu(self.fc4(h))
        z = z.view(-1, num_c3, 3, 3)
        z = F.relu(self.deconv3(z))
        z = F.relu(self.deconv1(z))
        z = self.deconv2(z)
        return h, z
