import model
import torch
import numpy as np
import matplotlib.pyplot as plt
import seaborn as sns
from torchvision import datasets, transforms
from sklearn.cluster import KMeans
from sklearn.metrics import confusion_matrix
from munkres import Munkres

def make_cost_matrix(c1, c2):
    """
    """
    uc1 = np.unique(c1)
    uc2 = np.unique(c2)
    l1 = uc1.size
    l2 = uc2.size
    assert(l1 == l2 and np.all(uc1 == uc2))

    m = np.ones([l1, l2])
    for i in range(l1):
        it_i = np.nonzero(c1 == uc1[i])[0]
        for j in range(l2):
            it_j = np.nonzero(c2 == uc2[j])[0]
            m_ij = np.intersect1d(it_j, it_i)
            m[i,j] =  -m_ij.size
    return m

def translate_clustering(clt, mapper):
    return np.array([ mapper[i] for i in clt ])

def accuracy(cm):
    """computes accuracy from confusion matrix"""
    return np.trace(cm, dtype=float) / np.sum(cm)


net = model.AE()
#net.load_state_dict(torch.load('./../result/Test_AE.pth', map_location = 'cpu'))
#net.load_state_dict(torch.load('./../result/AE_RAdam.pth', map_location= 'cpu'))
#net.load_state_dict(torch.load('./../result/AE_denoising_0.02.pth', map_location = 'cpu'))
#net.load_state_dict(torch.load('./../result/AE_scheduler.pth', map_location='cpu'))
#net.load_state_dict(torch.load('./paper/Test_AE_per9512.pth', map_location = 'cpu'))
#net.load_state_dict(torch.load('./../result/AE_joint_update.pth', map_location='cpu'))
net.load_state_dict(torch.load('./../result/AE_joint_Up_Ga001.pth', map_location = 'cpu'))

net.eval()

standarization = transforms.Compose([
                    transforms.ToTensor(),
                    transforms.Normalize(mean=(0.1307,),
                                        std=(0.3081,))])
test_data = datasets.MNIST('./../data/', train=False, transform=standarization, download=True)
batch_size = 10000
test_data_loader = torch.utils.data.DataLoader(test_data, batch_size, shuffle=False)

train_data = datasets.MNIST('./../data/', train=True, transform=standarization, download=True)
batch_size_train = 60000
train_data_loader = torch.utils.data.DataLoader(train_data, batch_size_train, shuffle=False)


for img, label in test_data_loader:
    hf, recon = net(img) #(10000, 10)

hf_num = hf.detach().numpy()
kmeans = KMeans(n_clusters=10, n_init=20).fit(hf_num)
prediction = kmeans.predict(hf_num)
label = label.detach().numpy()

cm = confusion_matrix(label, prediction)
ax = sns.heatmap(cm, annot=True, fmt="d", cmap="Blues")
#plt.savefig('./../result/image.png')

cost_matrix = make_cost_matrix(label, prediction)
m = Munkres()
indexes = m.compute(cost_matrix)
mapper = { old: new for (old, new) in indexes }

new_labels = translate_clustering(label, mapper)
new_cm = confusion_matrix(new_labels, prediction)
print(accuracy(new_cm))
plt.figure()
ax = sns.heatmap(new_cm, annot=True, fmt="d", cmap="Blues")
#plt.savefig('./../result/per_image.png')


for img, label in train_data_loader:
    hf, recon = net(img) #(10000, 10)

hf_num = hf.detach().numpy()
kmeans = KMeans(n_clusters=10, n_init=20).fit(hf_num)
prediction = kmeans.predict(hf_num)
label = label.detach().numpy()

cm = confusion_matrix(label, prediction)
ax = sns.heatmap(cm, annot=True, fmt="d", cmap="Blues")
#plt.savefig('./../result/train_image.png')

cost_matrix = make_cost_matrix(label, prediction)
m = Munkres()
indexes = m.compute(cost_matrix)
mapper = { old: new for (old, new) in indexes }

new_labels = translate_clustering(label, mapper)
new_cm = confusion_matrix(new_labels, prediction)
print("training accuracy : ", accuracy(new_cm))
plt.figure()
ax = sns.heatmap(new_cm, annot=True, fmt="d", cmap="Blues")
#plt.savefig('./../result/per_train_image.png')
print(kmeans.cluster_centers_)
a = np.array((10,10))
a = kmeans.cluster_centers_
np.save("./cluster.npy", a)
