import model
import torch
import torch.nn as nn
import torch.optim as optim
from torchvision import datasets, transforms
from torch.optim.lr_scheduler import StepLR
import time
import numpy as np
#net define

device_num = 1
net = model.AE()
net.load_state_dict(torch.load('./paper/Test_AE_per9512.pth', map_location = 'cpu'))
net.cuda(device_num) 

print(net)

# data load

standarization = transforms.Compose([
                    transforms.ToTensor(),
                    transforms.Normalize(mean=(0.1307,),
                                        std=(0.3081,))])
train_data = datasets.MNIST('./../data/', train=True, transform=standarization, download=True)

print('training data :', len(train_data))

batch_size = 256

train_data_loader = torch.utils.data.DataLoader(train_data, batch_size, shuffle=True)

# optimizer
criterion = nn.MSELoss()
optimizer = optim.Adam(net.parameters())
# training
num_epochs = 200

cluster_means = np.load("./cluster.npy")
cluster_means = torch.from_numpy(cluster_means).cuda(device_num)

def cluster_update(update_idx, tau_idx, cluster_means, hf):
    with torch.no_grad():
        for idx in update_idx:
            cluster_means[tau_idx[idx]] = (cluster_means[tau_idx[idx]] * 15 + hf[idx])/16 

def tau_comp(hf, cluster_means):
    loss = 0.0
    with torch.no_grad():
        thr = 0.8
        tau = torch.matmul(hf, cluster_means)
        tau_max, tau_idx = torch.max(tau, dim = 1) #values, indices
        tau_max[tau_max>thr] = 1
        update_idx = (tau_max==1).nonzero().view(-1)
    for idx in update_idx:
        loss += criterion(hf[idx], cluster_means[tau_idx[idx]]) 
    if(len(update_idx) == 0):
        cl_loss = 0.0
    else:
        cl_loss = loss/len(update_idx)
    return cl_loss, update_idx, tau_idx
gamma = 0.01

new_cluster = cluster_means

for epoch in range(num_epochs):
    start_time = time.time()
    running_loss = 0.0
    #new_cluster = cluster_means
    for img, label in train_data_loader:
        img = img.cuda(device_num)
        optimizer.zero_grad()
        hf, recon = net(img)
        cl_loss, up_idx, tau_idx = tau_comp(hf, new_cluster)
        recon_loss = criterion(img, recon)
        loss = cl_loss * gamma + recon_loss
        loss.backward()
        optimizer.step()
        running_loss += loss.item()
        cluster_update(up_idx, tau_idx, new_cluster, hf)
    #scheduler.step()
    end_time = time.time() 
    print('epoch: %2d loss: %.3f elpased time: %.3f' %(epoch+1, running_loss, end_time-start_time))

new_cluster = new_cluster.detach().numpy()
torch.save(net.state_dict(), './../result/AE_joint_Up_Ga001.pth')
np.save('new_cluster.npy', new_cluster)
